/***
 * Juliac
 * Copyright (C) 2010-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Frdric Loiret
 */

package org.objectweb.fractal.juliac.opt.ultramerge;

import java.io.IOException;
import java.lang.annotation.Annotation;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.api.SpoonSupportItf;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.desc.AttributeDesc;
import org.objectweb.fractal.juliac.desc.BindingDesc;
import org.objectweb.fractal.juliac.desc.ComponentDesc;
import org.objectweb.fractal.juliac.spoon.SpoonHelper;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;

import spoon.reflect.CoreFactory;
import spoon.reflect.Factory;
import spoon.reflect.code.CtAbstractInvocation;
import spoon.reflect.code.CtAssignment;
import spoon.reflect.code.CtBlock;
import spoon.reflect.code.CtExpression;
import spoon.reflect.code.CtFieldAccess;
import spoon.reflect.code.CtInvocation;
import spoon.reflect.code.CtLiteral;
import spoon.reflect.code.CtNewClass;
import spoon.reflect.code.CtReturn;
import spoon.reflect.code.CtStatement;
import spoon.reflect.code.CtStatementList;
import spoon.reflect.code.CtVariableAccess;
import spoon.reflect.cu.SourcePosition;
import spoon.reflect.declaration.CtAnnotation;
import spoon.reflect.declaration.CtAnonymousExecutable;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.declaration.CtConstructor;
import spoon.reflect.declaration.CtElement;
import spoon.reflect.declaration.CtExecutable;
import spoon.reflect.declaration.CtField;
import spoon.reflect.declaration.CtMethod;
import spoon.reflect.declaration.CtNamedElement;
import spoon.reflect.declaration.CtPackage;
import spoon.reflect.declaration.CtParameter;
import spoon.reflect.declaration.CtSimpleType;
import spoon.reflect.declaration.ModifierKind;
import spoon.reflect.factory.AnnotationFactory;
import spoon.reflect.factory.ClassFactory;
import spoon.reflect.factory.CodeFactory;
import spoon.reflect.factory.ConstructorFactory;
import spoon.reflect.factory.ExecutableFactory;
import spoon.reflect.factory.FieldFactory;
import spoon.reflect.factory.MethodFactory;
import spoon.reflect.factory.TypeFactory;
import spoon.reflect.reference.CtArrayTypeReference;
import spoon.reflect.reference.CtExecutableReference;
import spoon.reflect.reference.CtFieldReference;
import spoon.reflect.reference.CtTypeParameterReference;
import spoon.reflect.reference.CtTypeReference;
import spoon.reflect.reference.CtVariableReference;
import spoon.reflect.visitor.Query;
import spoon.reflect.visitor.filter.AbstractFilter;
import spoon.support.reflect.cu.SourcePositionImpl;

/**
 * Source code generator which merges all component business code in a single
 * class. This class is abstract in the sense that it is independent of the
 * annotation framework used for implementing component business code.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Frdric Loiret <Frederic.Loiret@inria.fr>
 * @since 2.2.4
 */
public abstract class AbstractUltraMerge {

	protected Juliac jc;
    protected UMClasses repo;
    
    public AbstractUltraMerge( Juliac jc ) {
    	this.jc = jc;
        repo = new UMClasses(jc);
    }
    
    /**
     * Generate the source code of a factory class for the specified component
     * descriptor.
     * 
     * @param adl         the fully-qualified name of the ADL descriptor
     * @param targetname  the fully-qualified name of the factory class
     */
    public void generate( ComponentDesc cdesc, String targetname )
    throws IOException {
        try {
            innergenerate(cdesc,targetname);
        }
        catch( IllegalComponentCodeException icce ) {
            /*
             * Should throw new IOException(e) but IOException(Throwable) is
             * only available since JDK 6 (and then fails from compiling with
             * JDK 5).
             */
            IOException ioe = new IOException();
            ioe.initCause(icce);
            throw ioe;
        }
    }
    
    private void innergenerate( ComponentDesc cdesc, String targetname )
    throws IOException, IllegalComponentCodeException {

        // Merge the business code
        List<CtSimpleType<?>> types = merge(cdesc,targetname);

        // Dump the generated types. Skip inner types.
        JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        for (CtSimpleType<?> type : types) {
        	if( type.isTopLevel() ) {
                spoon.generateSourceCode(type);
        	}
        }
    }    

    /**
     * Merge the business code of the specified component descriptor into a
     * single class. This method returns a list of types containing the merged
     * {@link CtClass} and all package protected types which are referenced by
     * the business code and which have been copied in the package of the merged
     * class (otherwise these types would be unreachable.)
     * 
     * @param cdesc       the component descriptor
     * @param targetname  the fully-qualified name of the merged class
     * @return            the created types
     */
    private List<CtSimpleType<?>> merge(
		ComponentDesc cdesc, String targetname )
    throws IllegalComponentCodeException {
                
        /*
         * Construct a model of the application by retrieving the {@link
         * UMClass}es referenced by the specified {@link ComponentDesc}.
         */
        repo.createAll(cdesc);
        
        /*
         * Check that the source classes can be merged.
         * Store the names of all component classes including their super
         * classes before flattening which discards the inheritance hierarchy.
         * These names will be used latter on when handling package protection.
         */
        mergeabilityCheck(cdesc);
        Set<String> allUMClassNames = repo.getAllUMClassNames();
        
        /*
         * Flatten the inheritance hierarchy.
         * Recreate a model of the application.
         * 
         * TODO rewrite to avoid having to recreate the model
         */
        flatten();
        repo.clear();
        repo.createAll(cdesc);
        
        /*
         * For each @Requires annotated field corresponding to a client
         * collection interface, create as many @Requires annotated fields
         * associated with a client singleton interface as bindings defined for
         * the collection interface.
         */
        createSingletonRequiresAnnotatedFieldsForCollections();
        
        /*
         * Create component attributes initialization statements.
         * Store binding information by processing @Requires annotated fields.
         */
        createComponentAttributeInitializationStatements();
        processRequiresAnnotatedFields();

        /*
         * Replace method calls on @Requires annotated fields by direct calls to
         * the method (since all methods will be merged in one unique class).
         */
        transformRequiresAnnotatedFieldAccesses();
        
        /*
         * Remove code elements unrelevant for the merged code.
         */
        removeUnrelevantCodeElements();
        
        /*
         * Merge the {@link CtClass}es.
         */
        CtClass<?> mixed = merge(targetname);
        mergeNoArgumentConstructors(mixed);
        
        /*
         * If the component descriptor provides a java.lang.Runnable/r server
         * interface and if there is not already a main(String[]) method in the
         * merged class, add one to launch the application.
         */
        addMainMethod(cdesc,mixed);
        
        /*
         * Handle package protected types and methods used by the merged class.
         */
        List<CtSimpleType<?>> stypes =
        	handlePackageProtection(mixed,allUMClassNames);
        
        /*
         * Add the merged class in created types if it hasn't been added
         * already.
         */
        if( ! stypes.contains(mixed) ) {
            stypes.add(mixed);
        }
        
        /*
         * Check whether the types use the Fractal API. Since the code is
         * merged, the use of the Fractal API is at best unrelevant, at most
         * faulty. However it may happen that the Fractal API is used for other
         * components than the merged ones. As both kinds of usages would be
         * rather difficult to differentiate, we raise a warning instead of an
         * error.
         */
        for (CtSimpleType<?> stype : stypes) {
        	Set<CtTypeReference<?>> trefs =
        		SpoonHelper.getUsedTypes(stype,false);
        	for (CtTypeReference<?> tref : trefs) {
				String qname = tref.getQualifiedName();
				if( qname.startsWith("org.objectweb.fractal.api.") ||
					qname.startsWith("org.objectweb.fractal.util.") ) {
					JuliacConfig jconf = jc.getJuliacConfig();
					Logger logger = jconf.getLogger();
					String msg = stype.getQualifiedName()+" uses "+qname;
					logger.warning(msg);
					break;  // reporting once per type is sufficient
				}
			}
		}
        
        return stypes;
    }
    

    // ----------------------------------------------------------------------
    // Abstract methods
    // ----------------------------------------------------------------------
    
    abstract protected String
        getRequiredAnnotatedUMFieldName( UMField<?> umfield );
    
    abstract protected boolean isRequiredAnnotatedUMField( UMField<?> umfield );

    abstract protected boolean
        isRequiredAnnotatedUMFieldCardinalitySingleton( UMField<?> umfield );

    abstract protected boolean
        isRequiredAnnotatedUMFieldCardinalityCollection( UMField<?> umfield );

    abstract protected void annotateRequiredCtField(
        AnnotationFactory af, CtField<?> field, String cltItfName );

    abstract protected boolean isRequiredAnnotation( Annotation a );

    /**
     * Return the setter method for the specified attribute name.
     */
    abstract protected UMMethod<?> getAttributeSetterUMMethod(
        UMClass<?> umclass, AttributeDesc adesc );
        
    abstract protected boolean
        isAttributeAnnotatedCtField( CtField<?> ctfield );

    abstract protected String
        getAttributeAnnotatedCtFieldName( CtField<?> ctfield );
    
    abstract protected CtTypeReference<?>
        getCollectionFieldType( UMField<?> umfield );
    
    /**
     * Return the class for variables which store references for collection
     * interfaces.
     */
    abstract protected Class<?> getCollectionClass();
    
    
    // ----------------------------------------------------------------------
    // Mergeability check
    // ----------------------------------------------------------------------
    
    /**
     * Check that the {@link CtClass}es of the architecture can be merged.
     * 
     * @param cdesc  the architecture descriptor
     * @throws IllegalComponentCodeException
     *      if the the {@link CtClass}es of the architecture can not be merged
     */
    protected void mergeabilityCheck( ComponentDesc cdesc )
    throws IllegalComponentCodeException {
        
        /*
         * Check the type of collection references.
         */
        Class<?> cl = getCollectionClass();
        if( !HashMap.class.equals(cl) && !ArrayList.class.equals(cl) ) {
            String msg =
                "Only HashMap and ArrayList are supported for variables "+
                "storing references to collection interfaces. Got: "+
                cl.getName();
            throw new IllegalComponentCodeException(msg);
        }
    }
    
    
    // ----------------------------------------------------------------------
    // Inheritance hierarchy flattening
    // ----------------------------------------------------------------------
    
    /**
     * Flatten the inheritance hierarchy of the {@link CtClass}es contained in
     * the architecture.
     */
    private void flatten() throws IllegalComponentCodeException {
        
        Collection<UMClass<?>> umclasses = repo.getAllUMClasses();
        
        /*
         * Iterate on all classes of the architecture.
         * Skip classes with no super class.
         */
        for (UMClass<?> umclass : umclasses) {
            CtClass<?> src = umclass.getCtClass();
            CtTypeReference<?> supercltref = src.getSuperclass();
            if( supercltref != null ) {
                // Merge the hierarchy into the component implementation class
                List<UMClass<?>> hierarchy = umclass.getUMSupers();
                merge(src,hierarchy);
                src.setSuperclass(null);
            }
        }
    }

    
    // ----------------------------------------------------------------------
    // Collection to singleton @Requires annoted fields for client interfaces
    // ----------------------------------------------------------------------
        
    /**
     * For each Requires annotated field corresponding to a client collection
     * interface, create as many Requires annotated fields associated with a
     * client singleton interface as bindings defined for the collection
     * interface.
     */
    private void createSingletonRequiresAnnotatedFieldsForCollections()
    throws IllegalComponentCodeException {
        
        Collection<UMClass<?>> umclasses = repo.getAllUMClasses();

        for ( UMClass<?> umclass : umclasses ) {
            
        	List<UMField<?>> umfields = umclass.getUMFields();
            List<CtField<?>> fields = new ArrayList<CtField<?>>();
            for (UMField<?> umfield : umfields) {
                if( isRequiredAnnotatedUMField(umfield) &&
                    isRequiredAnnotatedUMFieldCardinalityCollection(umfield) ) {
                	Map<String,CtField<?>> fieldsForSingleton =
                		processRequiresAnnotatedFieldCollection(umclass,umfield);
                	fields.addAll(fieldsForSingleton.values());
                }
            }
            
            // Add created fields in the umclass
            for (CtField<?> field : fields) {
				umclass.add(field);
			}
        }
    }
    
    /**
     * Process the specified Requires annotated field. The field corresponds to
     * a collection client interface. Add as many Requires annotated fields
     * associated with a client singleton interface as bindings defined for the
     * collection interface.
     */
    private Map<String,CtField<?>> processRequiresAnnotatedFieldCollection(
        UMClass<?> umclass, UMField<?> umfield )
    throws IllegalComponentCodeException {

        JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        TypeFactory tf = factory.Type();
        FieldFactory ff = factory.Field();
        AnnotationFactory af = factory.Annotation();
        CoreFactory coref = factory.Core();
        ClassFactory cf = factory.Class();
        
        /*
         * Get the name and the type of the client interface.
         */
        String refName = getRequiredAnnotatedUMFieldName(umfield);
        CtTypeReference<?> refType = getCollectionFieldType(umfield);

        /*
         * Retrieve the bindings for the current collection client interface.
         */
        ComponentDesc cdesc = umclass.getComponentDesc();
        List<BindingDesc> bdescs = new ArrayList<BindingDesc>();
        List<BindingDesc> allbdescs = cdesc.getBindingDescs();
        for (BindingDesc allbdesc : allbdescs) {
            String cltItfName = allbdesc.getCltItfName();
            if( cltItfName.startsWith(refName) ) {
                bdescs.add(allbdesc);
            }
        }
        
        /*
         * Create as many Requires annotated fields associated with a client
         * singleton interface as bindings defined for the collection interface.
         */
        Map<String,CtField<?>> fieldsForSingleton =
            new HashMap<String,CtField<?>>();
        CtClass<?> ctclass = umclass.getCtClass();        
        if( bdescs.size() > 0 ) {
            for (BindingDesc bdesc : bdescs) {
                String cltItfName = bdesc.getCltItfName();
                // TODO are there illegal characters other than minus in field names? 
                String fieldName = cltItfName.replace('-','_');
                CtField<?> fieldForSingleton = ctclass.getField(fieldName);
                if( fieldForSingleton != null ) {
                    // Duplicate field name
                    String msg =
                        "Trying to insert field "+fieldName+
                        ". A field with the same name already exists: "+
                        SpoonHelper.toEclipseClickableString(fieldForSingleton);
                    // TODO compute an unique name to insert the field
                    throw new IllegalComponentCodeException(msg);
                }
                Set<ModifierKind> mods = new HashSet<ModifierKind>();
                mods.add(ModifierKind.PRIVATE);
                fieldForSingleton = ff.create(ctclass,mods,refType,fieldName);
                annotateRequiredCtField(af,fieldForSingleton,cltItfName);
                fieldsForSingleton.put(cltItfName,fieldForSingleton);
            }
        }
        
        /*
         * Create the initialization expression for the @Requires annotated
         * field corresponding to the collection client interface.
         */
        
        // new HashMap or new ArrayList
        Class<?> collectionClass = getCollectionClass();
        CtNewClass newcl = coref.createNewClass();
        newcl.setType(tf.createReference(collectionClass));
        
        // anonymous sub class of HashMap/ArrayList
        CtClass<?> cl = cf.create("");
        cl.setParent(newcl);
        cl.setSuperclass(tf.createReference(collectionClass));
        newcl.setAnonymousClass(cl);
        
        // block of code for the anonymous sub class
        CtAnonymousExecutable anonexec = coref.createAnonymousExecutable();
        anonexec.setParent(cl);
        CtBlock<?> block = coref.createBlock();
        block.setParent(anonexec);
        List<CtStatementList<?>> statlist =
            new ArrayList<CtStatementList<?>>();
        block.setStatements((List)statlist);
        anonexec.setBody(block);
        List<CtAnonymousExecutable> anonexecs =
            new ArrayList<CtAnonymousExecutable>();
        anonexecs.add(anonexec);
        cl.setAnonymousExecutables(anonexecs);
        
        /*
         * Iterate on each binding defined in the collection client interface.
         */
        for (BindingDesc bdesc : bdescs) {
            createCollectionFieldInitializer(
                refType, bdesc, block, fieldsForSingleton );
        }
        
        /*
         * Initialization expression for the field which holds the Map
         * corresponding to the collection client interface.
         */
        CtField<?> field = umfield.getCtField();
        field.setDefaultExpression((CtExpression)newcl);
        newcl.setParent(field);
        
        /*
         * Remove the @Requires annotation on the field.
         * field.getAnnotations().remove(r) throws:
         *     java.lang.ClassCastException: $Proxy0 cannot be cast to java.lang.Comparable
         */
        Set<CtAnnotation<? extends Annotation>> annots = field.getAnnotations();
        Set<CtAnnotation<? extends Annotation>> newAnnots =
            new HashSet<CtAnnotation<? extends Annotation>>();
        for (CtAnnotation<? extends Annotation> annot : annots) {
            Annotation a = annot.getActualAnnotation();
            if( ! isRequiredAnnotation(a) ) {
                newAnnots.add(annot);
            }
        }
        field.setAnnotations(newAnnots);
        
        return fieldsForSingleton;
    }
    
    private void createCollectionFieldInitializer(
        CtTypeReference<?> refType, BindingDesc bdesc, CtBlock<?> block,
        Map<String,CtField<?>> fieldsForSingleton )
    throws IllegalComponentCodeException {

        JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        CoreFactory coref = factory.Core();
        ExecutableFactory ef = factory.Executable();
        ClassFactory cf = factory.Class();
        MethodFactory mf = factory.Method();

        /*
         * Add the corresponding element in the HashMap which initializes the
         * collection interface.
         */
        
        Class<?> collectionClass = getCollectionClass();
        
        // invocation to HashMap#put or ArrayList#add
        CtInvocation<?> inv = coref.createInvocation();
        inv.setParent(block);
        StringBuffer signature = new StringBuffer();
        signature.append(Object.class.getName());
        signature.append(' ');
        signature.append(HashMap.class.getName());
        signature.append(CtExecutable.EXECUTABLE_SEPARATOR);
        if( HashMap.class.equals(collectionClass) ) {
            signature.append("put");
            signature.append('(');
            signature.append(Object.class.getName());
            signature.append(',');            
        }
        else {  // ArrayList - checked by {@link #mergeabilityCheck}
            signature.append("add");
            signature.append('(');            
        }
        signature.append(Object.class.getName());
        signature.append(')');
        CtExecutableReference<?> eref =
            ef.createReference(signature.toString());
        inv.setExecutable((CtExecutableReference)eref);
        
        List<CtExpression<?>> arguments = new ArrayList<CtExpression<?>>();
        String cltItfName = bdesc.getCltItfName();

        // 1st argument for the call to put
        if( HashMap.class.equals(collectionClass) ) {
            CtLiteral<String> lit = coref.createLiteral();
            lit.setParent(inv);
            lit.setValue(cltItfName);
            arguments.add(lit);
        }
        
        // 2nd argument for the call to put / 1st argument for the call to add

        // new <type of the collection interface>
        CtNewClass<?> newcl = coref.createNewClass();
        newcl.setParent(inv);
        newcl.setType((CtTypeReference)refType);

        // anonymous sub class of <type of the collection interface>
        CtClass<?> cl = cf.create("");
        cl.setParent(newcl);
        cl.setSuperclass(refType);
        newcl.setAnonymousClass(cl);

        arguments.add(newcl);        
        inv.setArguments(arguments);
        List<CtStatement> statlist = block.getStatements();
        statlist.add(inv);
        
        /*
         * Iterate on all methods defined in the <type of the collection
         * interface>.
         * 
         * Do not use CtTypeReference#getAllExecutables().
         * See the Javadoc of SpoonHelper#getAllExecutables() for the rationale.
         */
        Collection<CtExecutableReference<?>> erefs =
            SpoonHelper.getAllExecutables(refType);
        if( erefs.size() == 0 ) {
        	String msg =
        		"Interfaces without any method are not supported for client" +
        		"collection interfaces: "+refType.getQualifiedName();
        	// TODO move this test to mergeabilityCheck()
        	throw new IllegalComponentCodeException(msg);
        }
        
        for (CtExecutableReference<?> eref2 : erefs) {
            if( eref2.isStatic() || eref2.isConstructor() ) {
                continue;
            }
            CtExecutable<?> exec = eref2.getDeclaration();
            if( ! (exec instanceof CtMethod<?>) ) {
                continue;
            }
            
            // Only public method from here
            CtMethod<?> origmeth = (CtMethod<?>) exec;
            List<CtParameter<?>> methParams = origmeth.getParameters();
            
            CtMethod<?> meth = mf.create(cl,origmeth,true);
            meth.getModifiers().add(ModifierKind.PUBLIC);
            cl.getMethods().add(meth);
            
            // body of the method
            CtBlock<?> body = coref.createBlock();
            body.setParent(meth);
            List<CtStatement> bodystatlist = new ArrayList<CtStatement>();
            body.setStatements(bodystatlist);
            meth.setBody((CtBlock)body);
            
            /*
             * Invocation to delegate the call to the field which corresponds to
             * the singleton client interface.
             */
            CtInvocation<?> inv2 = coref.createInvocation();
            inv2.setParent(body);
            inv2.setExecutable((CtExecutableReference)eref2);
            CtField<?> fieldForSingleton = fieldsForSingleton.get(cltItfName);
            CtFieldAccess<?> target = coref.createFieldAccess();
            target.setType((CtTypeReference)refType);
            target.setVariable((CtFieldReference)fieldForSingleton.getReference());
            target.setParent(inv2);
            inv2.setTarget(target);
            
            // invocation arguments
            List<CtExpression<?>> args2 = new ArrayList<CtExpression<?>>();
            for (CtParameter<?> param : methParams) {
                CtVariableAccess<?> va = coref.createVariableAccess();
                va.setParent(inv2);
                va.setVariable((CtVariableReference)param.getReference());
                args2.add(va);
            }
            inv2.setArguments(args2);
            
            // handle non void returning method
            CtStatement stat = null;
            CtTypeReference<?> rtype = origmeth.getType();
            if( rtype.getQualifiedName().equals("void") ) {
            	stat = inv2;
            }
            else {
            	CtReturn<?> ret = coref.createReturn();
            	ret.setParent(body);
            	ret.setReturnedExpression((CtExpression)inv2);
            	inv2.setParent(ret);
            	stat = ret;
            }
            
            bodystatlist.add(stat);
        }
    }
    
    
    // ----------------------------------------------------------------------
    // Constructor and attribute initializations
    // ----------------------------------------------------------------------
    
    /**
     * For all component attributes which need to be initialized, create the
     * corresponding initialization statements and insert them in a no argument
     * constructor.
     */
    private void createComponentAttributeInitializationStatements() {
            
        JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        CodeFactory codef = factory.Code();
        CoreFactory coref = factory.Core();
        
        Collection<UMClass<?>> umclasses = repo.getAllUMClasses();
        
        /*
         * Iterate on the no argument constructors and append the statements
         * needed to initialize the component attributes.
         */
        for (UMClass<?> umclass : umclasses) {
            CtClass<?> src = umclass.getCtClass();
            ComponentDesc cdesc = repo.get(umclass);
            
            /*
             * No argument constructor.
             */
            CtConstructor<?> ctr = src.getConstructor();
            CtBlock<?> body = (CtBlock<?>) ctr.getBody();  // cast needed by javac
            
            /*
             * Component attribute initializations.
             */
            Set<String> attributeNames = cdesc.getAttributeNames();
            for (String name : attributeNames) {
                
                AttributeDesc adesc = cdesc.getAttribute(name);
                String value = adesc.getValue();
                
                UMMethod<?> ummethod =
                    getAttributeSetterUMMethod(umclass,adesc);
                
                if( ummethod == null ) {
                    /*
                     * No setter method. Try to retrieve the field.
                     */
                    UMField<?> umfield =
                        getAttributeAnnotatedUMField(umclass,name);
                    CtField<?> ctfield = umfield.getCtField();
                    CtTypeReference<?> tref = ctfield.getType();
                    CtFieldReference<?> fref = ctfield.getReference();
                    
                    CtLiteral<?> lit = SpoonHelper.toCtLiteral(tref,value);
                    CtAssignment<?,?> assign =
                        codef.createVariableAssignment(fref,false,(CtLiteral)lit);
                    body.insertEnd(assign);

                    // Register this new field access
                    CtFieldAccess<?> fa = coref.createFieldAccess();
                    fa.setVariable((CtVariableReference)fref);
                    umfield.addCtFieldAccess((CtFieldAccess)fa);
                }
                else {
                    CtMethod<?> ctmethod = ummethod.getCtMethod();
                    List<CtParameter<?>> ctparams = ctmethod.getParameters();
                    CtParameter<?> ctparam = ctparams.get(0);
                    CtTypeReference<?> tref = ctparam.getType();
                    CtExecutableReference<?> eref = ctmethod.getReference();
                    
                    CtLiteral<?> lit = SpoonHelper.toCtLiteral(tref,value);
                    CtInvocation<?> inv = 
                        codef.createInvocation(null,eref,lit);
                    body.insertEnd(inv);
                    
                    // Register this new method invocation
                    ummethod.addCtInvocation((CtInvocation)inv);
                }
            }
        }
    }
    
    /**
     * Return the field which corresponds to the specified attribute name.
     */
    private UMField<?> getAttributeAnnotatedUMField(
        UMClass<?> umclass, String name )
    throws JuliacRuntimeException {
        
        List<UMField<?>> umfields = umclass.getUMFields();
        for (UMField<?> umfield : umfields) {
            
            /*
             * Check whether the field is @Attribute annotated.
             */
            CtField<?> ctfield = umfield.getCtField();
            if( isAttributeAnnotatedCtField(ctfield) ) {                
                String attrname = getAttributeAnnotatedCtFieldName(ctfield);
                if( attrname.equals(name) ) {
                    return umfield;
                }
            }
            
            /*
             * No @Attribute annotation. Check the name of the field.
             */
            String fname = umfield.getSimpleName();
            if( fname.equals(name) ) {
                return umfield;
            }
        }
        
        String msg =
            "No such component attribute "+name+" in "+
            umclass.getCtClass().getQualifiedName();
        throw new JuliacRuntimeException(msg);
    }
        
    
    // ----------------------------------------------------------------------
    // Process @Requires annotated fields and store binding information in
    // the model of the application.
    // ----------------------------------------------------------------------
    
    /**
     * Process the Requires annotated fields contained in the {@link CtClass}es
     * of the architecture.
     */
    private void processRequiresAnnotatedFields() {
        
        Collection<UMClass<?>> umclasses = repo.getAllUMClasses();

        for ( UMClass<?> umclass : umclasses ) {
            List<UMField<?>> umfields = getBoundReferenceFields(umclass);
            for (UMField<?> umfield : umfields) {
                processRequiresAnnotatedFieldSingleton(umclass,umfield);
            }
        }
    }
    
    /**
     * Process the specified Requires annotated field. The field corresponds to
     * a singleton client interface.
     */
    private void processRequiresAnnotatedFieldSingleton(
        UMClass<?> umclass, UMField<?> umfield )
    throws JuliacRuntimeException {

        /*
         * Get the name of the client interface.
         */
        String refName = getRequiredAnnotatedUMFieldName(umfield);
        
        /*
         * Work on a copy of field accesses to avoid
         * ConcurrentModificationException when removing the field accesses from
         * the list.
         */
        List<CtFieldAccess<?>> fas =
            new ArrayList<CtFieldAccess<?>>(umfield.getCtFieldAccesses());
        
        /*
         * Get the target class of the binding if there is one.
         */
        ComponentDesc cdesc = umclass.getComponentDesc();
        BindingDesc bdesc = cdesc.getBinding(refName);
        if( bdesc == null ) {
            CtField<?> field = umfield.getCtField();
            String msg =
                "No such client interface "+
                SpoonHelper.toEclipseClickableString(field)+
                " in component "+cdesc.getName();
            throw new JuliacRuntimeException(msg);
        }
        UMClass<?> target = getTargetComponentUMClass(bdesc);
        if( target == null ) {
            return;
        }

        /*
         * Register method invocations on @Requires annotated fields in
         * the corresponding UMMethod.
         */
        for (CtFieldAccess<?> fa : fas) {
            CtElement parent = fa.getParent();
            if( parent instanceof CtInvocation<?> ) {
                
                CtInvocation<?> inv = (CtInvocation<?>) parent;
                
                /*
                 * I bumped into a case where the field was one of the
                 * parameters of a method invocation, e.g. a.m(field), and where
                 * the parent element of the field was the invocation. I'm not
                 * sure whether this is normal or not.
                 * 
                 * Check that the target of the invocation is the field access.
                 * The target is null for invocations of a static method.
                 */
                CtExpression<?> invtarget = inv.getTarget();
                if( invtarget == null ) {
                    // Invocation of a static method
                    continue;
                }
                if( ! invtarget.equals(fa) ) {
                    continue;
                }

                CtExecutableReference<?> eref = inv.getExecutable();
                String invokedMethodName = eref.getSimpleName();
                List<CtTypeReference<?>> invokedMethodParameterTypes =
                    eref.getParameterTypes();
                
                UMMethod<?> invokedMethod =
                    target.getUMMethod(
                        invokedMethodName, invokedMethodParameterTypes );
                
                /*
                 * Defensive programming. This should no longer happen since the
                 * code documented by the previous comment has been added. We
                 * keep it anyway just in case.
                 */
                if( invokedMethod == null ) {
                    StringBuffer msg = new StringBuffer("No such method ");
                    msg.append(target.getCtClass().getQualifiedName());
                    msg.append('#');
                    msg.append(invokedMethodName);
                    msg.append('(');
                    boolean first = true;
                    for (CtTypeReference<?> tref : invokedMethodParameterTypes) {
                        if(first) {first=false;} else {msg.append(',');}
                        msg.append(tref.getQualifiedName());
                    }
                    msg.append(')');
                    throw new JuliacRuntimeException(msg.toString());
                }
                
                invokedMethod.addCtInvocation((CtInvocation)inv);
                umfield.removeCtFieldAccess((CtFieldAccess)fa);
            }
        }
    }
    
    /**
     * Follow the path from the specified {@link BindingDesc} to the bound
     * primitive component (traverse composite components if needed) and return
     * the {@link UMClass} implementing the primitive component. Return
     * <code>null</code> if there is no bound component.
     */
    private UMClass<?> getTargetComponentUMClass( BindingDesc bdesc ) {
        ComponentDesc srv = bdesc.getSrv();
        while( srv.getContentClassName() == null ) {
            String srvItfName = bdesc.getSrvItfName();
            BindingDesc srvBdesc = srv.getBinding(srvItfName);
            if( srvBdesc == null ) {
                /*
                 * No bound component.
                 * 
                 * This happens for example with Cosmos where the push-notif-out
                 * collection interface on a dreamComposite has no bound
                 * component (because the top of the hierarchy is reached?).
                 * This may also happen with optional interfaces?
                 */
                return null;
            }
            bdesc = srvBdesc;
            srv = bdesc.getSrv();
        }
        UMClass<?> target = repo.get(srv);
        return target;
    }
    

    // ----------------------------------------------------------------------
    // Replace method calls on @Requires annotated fields by direct calls to
    // the method (since all methods will be merged in one unique class.)
    // ----------------------------------------------------------------------
    
    /**
     * Transform method invocations on Requires annotated field accesses
     * appearing in the {@link CtClass}es of the architecture to local method
     * invocations.
     */
    private void transformRequiresAnnotatedFieldAccesses() {

        JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        ExecutableFactory ef = factory.Executable();
        CoreFactory coref = factory.Core();
        
        Collection<UMClass<?>> umclasses = repo.getAllUMClasses();

        for( UMClass<?> umclass : umclasses ) {
            
            /*
             * Process component method invocations by setting the target
             * reference of the invocation to this.
             */
            List<UMMethod<?>> ummethods = umclass.getUMMethods();
            for (UMMethod<?> ummethod : ummethods) {
                
                List<CtInvocation<?>> invs = (List) ummethod.getCtInvocations();
                for (CtInvocation<?> inv : invs) {
                    
                    CtExecutableReference<?> invocationExecutable = inv.getExecutable();
                    CtClass<?> src = umclass.getCtClass();
                    CtTypeReference<?> declaringType = src.getReference();
                    CtTypeReference<?> type = inv.getType();
                    String methodName = invocationExecutable.getSimpleName();
                    List<CtTypeReference<?>> parameterTypes =
                        invocationExecutable.getParameterTypes();
                    
                    CtExecutableReference<?> eref =
                        ef.createReference(
                            declaringType,type,methodName,parameterTypes);
                    inv.setExecutable((CtExecutableReference)eref);

                    /*
                     * Set the target to this.
                     * For that, we should normally have to do "only":
                     *      inv.setTarget(null);
                     * 
                     * Yet, when collection interfaces come into play, it may
                     * happen that the following code is generated:
                     *     @Requires(cardinality = Cardinality.COLLECTION, name = "s")
                     *     private Map<String,Service> m = new HashMap() {{
                     *          put("s1" ,new Service() {
                     *              public void print(String msg) {
                     *                  print(msg);
                     *
                     * The print method which is invoked should be the one of
                     * the enclosing type, not the one which is declared in the
                     * anonymous subclass of Service.
                     * 
                     * To prevent this from happening, we generate:
                     *                  HelloWorld.this.print(msg);
                     * 
                     * which leads to the invocation of the correct print method
                     * (assuming HelloWorld is the enclosing type.)
                     */
                    CtMethod<?> enclosingMethod = inv.getParent(CtMethod.class);
                    if( enclosingMethod == null ) {
                        // May be a constructor
                        inv.setTarget(null);                        
                    }
                    else {
                        String enclosingMethodName = enclosingMethod.getSimpleName();
                        if( enclosingMethodName.equals(methodName) ) {
                            CtFieldAccess<?> fa = coref.createFieldAccess();
                            fa.setParent(inv);
                            fa.setType((CtTypeReference)declaringType);
                            CtFieldReference<?> fr = coref.createFieldReference();
                            fr.setDeclaringType((CtTypeReference)declaringType);
                            fr.setType((CtTypeReference)declaringType);
                            fr.setSimpleName("this");
                            fa.setVariable((CtVariableReference)fr);
                            inv.setTarget(fa);
                        }
                        else {
                            inv.setTarget(null);
                        }
                    }
                }
            }
            
            /*
             * Process accesses to @Requires annotated fields by setting the
             * target reference to this.
             */
            List<UMField<?>> umfields = getBoundReferenceFields(umclass);
            for (UMField<?> umfield : umfields) {
                List<CtFieldAccess<?>> fas =
                    (List) umfield.getCtFieldAccesses();
                for (CtFieldAccess<?> fa : fas) {
                    CtFieldReference<?> fref = fa.getVariable();
                    fref.setSimpleName("this");
                }
            }
        }
    }

    
    // ----------------------------------------------------------------------
    // Removals
    // ----------------------------------------------------------------------
    
    /**
     * Remove code elements unrelevant for the merged code.
     */
    protected void removeUnrelevantCodeElements() {
        Collection<UMClass<?>> umclasses = repo.getAllUMClasses();
        for (UMClass<?> umclass : umclasses) {
            removeRequiresAnnotatedFieldSingleton(umclass);
            removeOverrideAnnotations(umclass);
        }
    }
    
    /**
     * Remove @{@link Requires} annotated fields corresponding to singleton
     * client interfaces. Keep the ones which correspond to collection client
     * interfaces and to unbound interfaces.
     */
    private void removeRequiresAnnotatedFieldSingleton(
        UMClass<?> umclass ) {
        
        // Work on a copy to avoid ConcurrentModificationException
        List<UMField<?>> umfields = umclass.getUMFields();        
        List<UMField<?>> copy = new ArrayList<UMField<?>>();
        copy.addAll(umfields);

        for (UMField<?> umfield : copy) {
            if( isRequiredAnnotatedUMField(umfield) &&
                isRequiredAnnotatedUMFieldCardinalitySingleton(umfield) &&
                isRequiredInterfaceBound(umfield,umclass) ) {
                umfields.remove(umfield);
            }
        }        
    }

    /**
     * Remove the @{@link Override} annotations associated with the methods of
     * the specified class.
     */
    private void removeOverrideAnnotations( UMClass<?> umclass ) {
        
        CtClass<?> ctclass = umclass.getCtClass();
        Set<CtMethod<?>> methods = ctclass.getAllMethods();
        for ( CtMethod<?> method : methods ) {
            
            Set<CtAnnotation<?>> annots = method.getAnnotations();
            for ( CtAnnotation<?> annot : annots ) {
                CtTypeReference<?> tref = annot.getAnnotationType();
                String qname = tref.getQualifiedName();
                if( qname.equals( Override.class.getName() ) ) {
                    annots.remove(annot);
                }
            }
        }
    }
    
    
    // ----------------------------------------------------------------------
    // Class merging
    // ----------------------------------------------------------------------
    
    /**
     * Create a new {@link CtClass} named <code>targetClassName</code> by
     * merging existing {@link CtClass}es.
     * 
     * @param targetClassName  the name of the created class
     * @return                 the resulting merged class
     */
    private CtClass<?> merge( String targetClassName )
    throws IllegalComponentCodeException {
        
        JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        
        /*
         * Create the target class.
         */
        CtClass<?> target = factory.Class().create(targetClassName);
        target.setVisibility(ModifierKind.PUBLIC);
        
        /*
         * Merge classes.
         */
        Collection<UMClass<?>> umclasses = repo.getAllUMClasses();
        List<UMClass<?>> list = new ArrayList<UMClass<?>>();
        list.addAll(umclasses);
        merge(target,list);
        
        return target;
    }
    
    /**
     * Merge the specified {@link UMClass}es in the specified {@link CtClass}.
     * 
     * @throws IllegalComponentCodeException
     *      if one of the classes to be merged extends a super class
     */
    private static void merge(
        CtClass<?> target, Collection<UMClass<?>> umclasses )
    throws IllegalComponentCodeException {
        
        for (UMClass<?> umclass : umclasses) {
            
            /*
             * Check that the class to be merged has no super class.
             */
            CtClass<?> ctclass = umclass.getCtClass();
            CtTypeReference<?> superclass = ctclass.getSuperclass();
            if( superclass != null ) {
                String msg =
                    "Can not merge a class which extends another one: "+
                    SpoonHelper.toEclipseClickableString(ctclass)+" extends "+
                    superclass.getQualifiedName();
                throw new IllegalComponentCodeException(msg);
            }            
            
            /*
             * Merge fields.
             */
            List<UMField<?>> umfields = umclass.getUMFields();
            for (UMField<?> umfield : umfields) {                
                CtField<?> field = umfield.getCtField();
                String name = field.getSimpleName();
                CtField<?> f = target.getField(name);
                
                if( f != null ) {
                    // Duplicate field name
                    String umclassname =
                    	umclass.getCtClass().getQualifiedName();
                    String newName = name+'_'+umclassname.replace('.','_');
                    umfield.setSimpleName(newName);
                }
                
                // Reference the source location of the field
                String comment = field.getDocComment();
                if(comment==null) comment="";
                comment += SpoonHelper.toSeeLink(field)+"\n";
                field.setDocComment(comment);
                
                // Insert the field in the target class
                target.getFields().add(field);
                field.setParent(target);
            }
            
            /*
             * Merge methods.
             */
            List<UMMethod<?>> ummethods = umclass.getUMMethods();
            for (UMMethod<?> ummethod : ummethods) {
                
            	CtMethod<?> method = ummethod.getCtMethod();
                Set<ModifierKind> mods = method.getModifiers();
                if( mods.contains(ModifierKind.ABSTRACT) ) {
                	// Skip abstract methods
                	continue;
                }
                
                String name = method.getSimpleName();
                CtTypeReference<?>[] parameterTypes =
                	SpoonHelper.getParameterTypes(method);
                CtMethod<?> m = target.getMethod(name,parameterTypes);
                if( m != null ) {
                    // Duplicate method name
                    String umclassname =
                    	umclass.getCtClass().getQualifiedName();
                    String newName = name+'_'+umclassname.replace('.','_');
                    ummethod.setSimpleName(newName);
                    ummethod.setParent(target);
                }
                
                // Reference the source location of the method
                String comment = method.getDocComment();
                if(comment==null) comment="";
                comment += SpoonHelper.toSeeLink(method)+"\n";      
                method.setDocComment(comment);
                
                // Insert the method in the target class
                target.getMethods().add(method);
                method.setParent(target);
            }
            
            /*
             * Add implemented interfaces.
             */
            Set<CtTypeReference<?>> implementedInterfaces =
            	ctclass.getSuperInterfaces();
            target.getSuperInterfaces().addAll(implementedInterfaces);

            /*
             * Add inner types.
             */
            Set<CtSimpleType<?>> innertypes = ctclass.getNestedTypes();
            for (CtSimpleType<?> innertype : innertypes) {
                
            	String name = innertype.getSimpleName();
                CtSimpleType<?> t = target.getNestedType(name);
                if( t != null ) {
                    // Duplicate inner type
                    String msg = "Duplicate inner type: "+name;
                    // TODO to be implemented
                    throw new IllegalComponentCodeException(msg);
                }

                // Reference the source location of the inner type
                String comment = innertype.getDocComment();
                if(comment==null) comment="";
                comment += SpoonHelper.toSeeLink(innertype)+"\n";      
                innertype.setDocComment(comment);
                
                // Insert the inner type in the target class
                target.getNestedTypes().add(innertype);
                innertype.setParent(target);
            }            
        }
        
        /*
         * Update type references from the source classes to the merged one.
         * Perform this only after all the classes have been merged into target
         * to avoid missing dependencies (forward references to classes which
         * haven't been merged yet.)
         */
        CtTypeReference<?> dstTRef = target.getReference();
        for (UMClass<?> umclass : umclasses) {
            CtClass<?> ctclass = umclass.getCtClass();
            CtTypeReference<?> srcTRef = ctclass.getReference();
            SpoonHelper.updateTypeReferences(target,srcTRef,dstTRef);
        }
    }

    /**
     * Merge the statements of the no argument constructors contained in the
     * {@link CtClass}es of the architecture and insert the result in the no
     * argument constructor of the specified target {@link CtClass}.
     */
    private void mergeNoArgumentConstructors( CtClass<?> target ) {
        
        // TODO merge with previous method
        // TODO handle duplicate local variables
        
        JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        ConstructorFactory cf = factory.Constructor();
        CodeFactory codef = factory.Code();
        CoreFactory coref = factory.Core();
        CtTypeReference<?> dstTRef = target.getReference();
        
        Collection<UMClass<?>> umclasses = repo.getAllUMClasses();

        /*
         * Iterate on the no argument constructors and prepare a body by
         * concatening their statements.
         */
        CtBlock<?> body = coref.createBlock();
        for (UMClass<?> umclass : umclasses) {
            CtClass<?> src = umclass.getCtClass();
            CtTypeReference<?> srcTRef = src.getReference();
            
            /*
             * No argument constructor statement list.
             */
            CtConstructor<?> ctr = src.getConstructor();
            CtBlock<?> b = (CtBlock<?>) ctr.getBody();  // cast needed by javac
            CtStatementList<?> stats = codef.createStatementList(b);
            SpoonHelper.updateTypeReferences(stats,srcTRef,dstTRef);
            body.insertEnd(stats);
        }
        
        /*
         * Create a docComment to document where the merged constructors come
         * from.
         */
        String docComment = "";
        for (UMClass<?> umclass : umclasses) {
            CtClass<?> src = umclass.getCtClass();
            String name = src.getQualifiedName();
            String simpleName = src.getSimpleName();
            if( docComment.length() != 0 ) {
                docComment += "\n";
            }
            docComment += " @see "+name+"#"+simpleName+"()";
        }
        
        /*
         * Create a no argument constructor and insert it in target.
         */
        Set<ModifierKind> mks = new HashSet<ModifierKind>();
        mks.add(ModifierKind.PUBLIC);
        List<CtParameter<?>> parameters = new ArrayList<CtParameter<?>>();
        Set<CtTypeReference<? extends Throwable>> thrownTypes =
            new HashSet<CtTypeReference<? extends Throwable>>();
        CtConstructor ctr = cf.create(target,mks,parameters,thrownTypes);
        ctr.setBody((CtBlock)body);
        ctr.setDocComment(docComment);
        
        target.getConstructors().add(ctr);
    }

    
    // ----------------------------------------------------------------------
    // Add main method
    // ----------------------------------------------------------------------
    
    /**
     * If the specified component descriptor provides a
     * <code>java.lang.Runnable/r</code> server interface and if there is not
     * already a <code>main(String[])</code> method in the merged class, add one
     * to launch the application.
     * 
     * @param cdesc    the component descriptor
     * @param ctclass  the merged class
     */
    private void addMainMethod( ComponentDesc cdesc, CtClass<?> ctclass ) {
    
        /*
         * Check whether the component descriptor provides a
         * java.lang.Runnable/r server interface.
         */
        ComponentType ct = cdesc.getCT();
        try {
            InterfaceType it = ct.getFcInterfaceType("r");
            if( it.isFcClientItf() ) {
                // Not a server interface. Do nothing.
                return;
            }
            String signature = it.getFcItfSignature();
            if( ! signature.equals(Runnable.class.getName()) ) {
                // Not a java.lang.Runnable interface. Do nothing.
                return;
            }
        }
        catch (NoSuchInterfaceException e) {
            // No r interface. Do nothing.
            return;
        }
        
        Factory factory = ctclass.getFactory();
        TypeFactory tf = factory.Type();
        MethodFactory mf = factory.Method();
        CoreFactory coref = factory.Core();
        CodeFactory codef = factory.Code();

        /*
         * Check whether there is already a main(String[]) method.
         */
        CtArrayTypeReference<?> trefstringarray =
            tf.createArrayReference(String.class.getName());
        CtMethod<?> method = ctclass.getMethod("main",trefstringarray);
        if( method != null ) {
            // A main(String[]) already exists. Do nothing.
            return;
        }
        
        /*
         * Create the main(String[] args) method.
         */
        Set<ModifierKind> mods = new HashSet<ModifierKind>();
        mods.add(ModifierKind.PUBLIC);
        mods.add(ModifierKind.STATIC);
        CtTypeReference<?> rtref = tf.createReference("void");
        List<CtParameter<?>> params = new ArrayList<CtParameter<?>>();
        Set<CtTypeReference<? extends Throwable>> thrownTypes =
            new HashSet<CtTypeReference<? extends Throwable>>();
        CtBlock body = coref.createBlock();
        List<CtStatement> statements = new ArrayList<CtStatement>();
        body.setStatements(statements);
        method = mf.create(ctclass,mods,rtref,"main",params,thrownTypes,body);
        mf.createParameter(method,trefstringarray,"args");
        
        // Create the statement: Runnable r = new _ctclass_()
        String statstr =
        	"java.lang.Runnable r = new "+ctclass.getQualifiedName()+"()";
        CtStatement stat = codef.createCodeSnippetStatement(statstr);
        stat.setParent(body);
        statements.add(stat);

        // Create the statement: r.run()
        statstr = "r.run()";
        stat = codef.createCodeSnippetStatement(statstr);
        stat.setParent(body);
        statements.add(stat);
    }
    
    
    // ----------------------------------------------------------------------
    // Package protected types and methods handling
    // ----------------------------------------------------------------------
    
    /**
     * Copy package protected types and methods used by the specified merged
     * class in the package of this class. These elements would otherwise be
     * unreachable in this class.
     * 
     * @param src  the merged class
     * @param allUMClassNames
     * 			the names of original component classes including the names of
     * 			their super classes
     * @return  the list of copied types
     */
    private List<CtSimpleType<?>> handlePackageProtection(
    	CtClass<?> src, Set<String> allUMClassNames ) {
        
        List<CtSimpleType<?>> stypes = new ArrayList<CtSimpleType<?>>();
        
        JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        TypeFactory tf = factory.Type();
        
        /*
         * Qualified names of package protected referenced types. Store names
         * instead of {@link CtTypeReference}s in order to avoid unecessary
         * duplicate {@link CtTypeReference}s which reference the same type.
         */
        final List<String> qnames = new ArrayList<String>();
        handlePackageProtectedTypes(src,qnames);
        handlePackageProtectedExecutables(src,qnames);
        handlePackageProtectionTransitiveClosure(src,qnames);

        /*
         * Recursively process all package protected types and methods.
         * 
         * Don't use an iterator as the list may grow while iterating which
         * would throw a ConcurrentModificationException.
         */
        for (int i = 0; i < qnames.size(); i++) {

            String qname = qnames.get(i);
            CtSimpleType<?> stype = null;
        	
        	/*
        	 * Check whether qname corresponds to one of the component
        	 * implementations which have been merged. If this is the case,
        	 * replace it with the merged type.
        	 */
        	if( allUMClassNames.contains(qname) ) {
        		stype = src;
        	}
        	else {
            	CtTypeReference<?> tref = tf.createReference(qname);
                stype = tref.getDeclaration();
                handlePackageProtectedTypes(stype,qnames);
                handlePackageProtectedExecutables(stype,qnames);
                handlePackageProtectionTransitiveClosure(stype,qnames);        		
        	}
        	
            stypes.add(stype);            
        }
        
        /*
         * Iterate on package protected types to make them public and to move
         * them in the same package as src. Skip duplicate entries in stypes
         * which are due to merged types.
         */
        Set<CtSimpleType<?>> ustypes = new HashSet<CtSimpleType<?>>();
        ustypes.addAll(stypes);
        CtPackage pack = src.getPackage();
        for (CtSimpleType<?> stype : ustypes) {

        	// Skip inner types.
            if( ! stype.isTopLevel() ) {
            	continue;
            }
            
            if( stype.equals(src) ) {
            	// Skip src since it does not have to be moved. Will be created.
            	continue;
            }

        	/*
        	 * This block is a hack to force Spoon to output all types.
        	 * 
        	 * TODO fix compilation unit management associated to moved types
        	 */
            SourcePositionImpl sp = (SourcePositionImpl) stype.getPosition();
            if( sp != null ) {
                sp.setCompilationUnit(null);
            }
            
            // Reference the source location of the type
            String comment = stype.getDocComment();
            if(comment==null) comment="";
            comment += SpoonHelper.toSeeLink(stype)+"\n";      
            stype.setDocComment(comment);
            
            // Rename the type in case of duplicate type names in src
            String sname = stype.getSimpleName();
            if( containsSimpleName(pack,sname) ) {
            	CtPackage p = stype.getPackage();
            	String pname = p.getQualifiedName();
            	pname = pname.replace('.','_');
                sname = pname+'_'+stype.getSimpleName();
            	stype.setSimpleName(sname);
            }
            
            // Insert the type in the same package as src
            pack.getTypes().add(stype);
            stype.setParent(pack);
        }
        
        /*
         * Update type references to moved types in src and in all moved types.
         */
        for (int i = 0; i < qnames.size(); i++) {
        	
        	String qname = qnames.get(i);
        	CtSimpleType<?> stype = stypes.get(i);
        	CtTypeReference<?> tref = tf.createReference(qname);
            CtTypeReference<?> clonetref = stype.getReference();
            
            SpoonHelper.updateTypeReferences(src,tref,clonetref);            
            for (CtSimpleType<?> other : stypes) {
                SpoonHelper.updateTypeReferences(other,tref,clonetref);
			}
		}
        
        /*
         * Update type references in all moved types from the types which have
         * been merged in src to src.
         */
        CtTypeReference<?> clonetref = src.getReference();
        for (int i = 0; i < qnames.size(); i++) {
        	CtSimpleType<?> stype = stypes.get(i);
            for (String umclassname : allUMClassNames) {
            	CtTypeReference<?> tref = tf.createReference(umclassname);
                SpoonHelper.updateTypeReferences(stype,tref,clonetref);            
			}
		}

        /*
         * Return ustypes as a list. Do not return stypes to avoid duplicate
         * entries.
         */
        List<CtSimpleType<?>> lstypes = new ArrayList<CtSimpleType<?>>();
        lstypes.addAll(ustypes);
        
        return lstypes;
    }
    
    /**
     * Return <code>true</code> if the specified package contains a type with
     * the specified simple name.
     * 
     * @param pack   the package
     * @param sname  the simple name
     * @return	<code>true</code> if the specified package contains a type with
     * 			the specified simple name
     */
    private static boolean containsSimpleName( CtPackage pack, String sname ) {
    	Set<CtSimpleType<?>> stypes = pack.getTypes();
    	for (CtSimpleType<?> stype : stypes) {
			String name = stype.getSimpleName();
			if( name.equals(sname) ) {
				return true;
			}
		}
    	return false;
    }
    
    /**
     * Add to the specified list, the names of the package protected types
     * referenced by the specified {@link CtSimpleType} and not already
     * contained in the specified list.
     */
    private void handlePackageProtectedTypes(
        final CtSimpleType<?> src, final List<String> qnames ) {
        
    	Set<CtTypeReference<?>> trefs = SpoonHelper.getUsedTypes(src,false);
    	for (CtTypeReference<?> tref : trefs) {

    		String qname = tref.getQualifiedName();
    		if( qnames.contains(qname) ) {
    			// We have already encountered this type
    			continue;
    		}
            
    		// Skip generic types
            if( tref instanceof CtTypeParameterReference ) {
                continue;
            }
            
            // Skip non package protected types
    		UnifiedClass uclass = jc.create(qname);
            int mods = uclass.getModifiers();
            if( Modifier.isPublic(mods) || Modifier.isProtected(mods) ||
                Modifier.isPrivate(mods) ) {
                continue;
            }
            
            // Handle inner types
            UnifiedClass enclosing = uclass.getDeclaringClass();
            if( enclosing != null ) {
                // uclass is an inner type
                if( enclosing.isInterface() ) {
                    // By default all inner elements of an interface are public
                    continue;
                }
                String srcqname = src.getQualifiedName();
                String eqname = enclosing.getName();
                if( srcqname.equals(eqname) ) {
                	// Skip inner types defined in the source type
                	continue;
                }
            }
            
            /*
             * At this point of the execution, we have skipped all cases which
             * do not match. We can add the current type in the list of package
             * protected types. 
             */
            qnames.add(qname);
		}
    }

    /**
     * Add to the specified list, the names of the types containing invocations
     * of package protected methods and constructors referenced by the specified
     * {@link CtSimpleType} and not already contained in the specified list.
     */
    private void handlePackageProtectedExecutables(
		CtSimpleType<?> src, final List<String> qnames ) {

        Query.getElements(
            src,
            /*
             * Use CtAbstractInvocation to encompass CtInvocation and
             * CtNewClass. Both are needed since we must retrieve invocations of
             * package protected methods and constructors.
             */
            new AbstractFilter<CtAbstractInvocation<?>>(CtAbstractInvocation.class) {
                public boolean matches( CtAbstractInvocation<?> inv ) {
                    
                    CtExecutableReference<?> er = inv.getExecutable();
                    
                    if( er == null ) {
                    	/*
                    	 * When visiting a piece of code which has been
                    	 * generated by processRequiresAnnotatedFieldCollection
                    	 * for client collection interfaces, the model contains
                    	 * CtNewClass instances of anonymous inner classes where
                    	 * the executable reference has not been set.
                    	 * 
                         * Although this is a bug to fix, we skip it here,
                         * finger crossing that it won't concern other package
                         * protected executables :-(
                         * 
                    	 * TODO fix the generated part of the model
                    	 */
                    	return false;
                    }
                    
                    try {
                    	Object actual =
                    		er.isConstructor() ?
            				er.getActualConstructor() :
        					er.getActualMethod();
                        if( actual == null ) {
                            /*
                             * getModifiers() invokes getActualMethod().
                             * 
                             * It may happen that if the actual method (e.g.
                             * java.util.Vector#add(Object)) can not be
                             * retrieved because the invocation is performed
                             * with subtypes (e.g. v.add(aString)),
                             * getModifiers() returns an empty set instead of
                             * failing.
                             * 
                             * Although this is a bug to fix, we skip it here,
                             * finger crossing that it won't concern package
                             * protected types :-(
                             * 
                             * TODO fix declarations
                             */
                            return false;
                        }
                    }
                    catch( RuntimeException re ) {
                        Throwable cause = re.getCause();
                        if( cause instanceof ClassNotFoundException ) {
                            /*
                             * The issue is with the getModifiers() invocation
                             * (below) which returns an empty set if the actual
                             * method can not be retrieved.
                             * 
                             * It may happen that a ClassNotFoundException is
                             * thrown when getModifiers() invokes
                             * getActualMethod() which in its turn invokes
                             * getActualClass().
                             * 
                             * getActualMethod() compares the methods declared
                             * in the type by testing the strict equality of
                             * parameter types with getActualMethod().
                             * 
                             * The case may occur (e.g. with Cosmos) where the
                             * parameters are subtypes of these types and that
                             * the subtype declaration can not be retrieved. In
                             * theory the declaration is present. Has it been
                             * lost somewhere in the transformation?
                             * 
                             * Although this is a bug to fix, we skip it here,
                             * finger crossing that it won't concern package
                             * protected types :-(
                             * 
                             * TODO fix declarations
                             */
                            return false;
                        }
                        
                        // Other cases: rethrow the exception
                        throw re;
                    }
                    
                    /*
                     * Skip public and private methods, although we shouldn't
                     * encounter invocations to private methods, should we?
                     * 
                     * Do not skip protected methods, since they must be treated
                     * as package protected. Indeed, any invocation to a
                     * protected method in the same package is legal. Hence the
                     * class defining the protected method must be treated as
                     * the classes defining package protected methods.
                     */
                    Set<ModifierKind> mods = er.getModifiers();
                    if( mods.contains(ModifierKind.PUBLIC) ||
                		mods.contains(ModifierKind.PRIVATE) ) {
                        return false;
                    }
                    CtTypeReference<?> declaringType = er.getDeclaringType();
                    
                    /*
                     * Skip method defined in interfaces since they are by
                     * default public.
                     */
                	String qname = declaringType.getQualifiedName();
                    UnifiedClass uc = jc.create(qname);
                    if( uc.isInterface() ) {
                    	return false;
                    }

                	if( qnames.contains(qname) ) {
                    	// We have already encountered this type
                    	return false;
                	}

                    /*
                     * At this point of the execution, we have skipped all cases
                     * which do not match. We can add the current type in the
                     * list of package protected types. 
                     */
                    qnames.add(qname);

                    /*
                     * Always return false. We are not interested by the result
                     * of the query. Only the updating of trefs matters.
                     */
                    return false;
                }
            }
        );
    }        

    /**
     * Add to the specified list, the types used by the specified {@link
     * CtSimpleType} and referencing a type contained in the specified list.
     */
    private void handlePackageProtectionTransitiveClosure(
		CtSimpleType<?> src, final List<String> qnames ) {

    	Set<CtTypeReference<?>> trefs = SpoonHelper.getUsedTypes(src,true);
    	for (CtTypeReference<?> tref : trefs) {
			
    		String qname = tref.getQualifiedName();
    		if( qnames.contains(qname) ) {
            	// We have already encountered this type
    			continue;
    		}
    		
    		/*
    		 * If we can not get the declaration, this means that the source of
    		 * this type has not been parsed by Spoon. Most probably the type is
    		 * in bytecode form in the class path. We assume that in this case,
    		 * there is few chances that this type uses one of the types which
    		 * will be moved due to package protection.
    		 */
    		CtSimpleType<?> stype = tref.getDeclaration();
			if( stype == null ) {
				continue;
			}
			
	    	Set<CtTypeReference<?>> t2refs =
	    		SpoonHelper.getUsedTypes(stype,true);
			for (CtTypeReference<?> t2ref : t2refs) {
				String q2name = t2ref.getQualifiedName();
				if( qnames.contains(qname) ) {
					/*
					 * Needed even if the containment has been tested by the
					 * outter loop. The element may be added latter on by the
					 * next statements. Hence we need to test again the
					 * containment here to avoid duplicate insertions.
					 */
					continue;
				}
				if( qnames.contains(q2name) ) {
					qnames.add(qname);
				}
			}
		}
    	
    }
    
    
    // ----------------------------------------------------------------------
    // Utility methods
    // ----------------------------------------------------------------------
    
    /**
     * Return a stringified indication of the source position for the specified
     * {@link CtAnnotation}.
     */
    protected String toPositionString( CtAnnotation<?> annot ) {
        CtNamedElement ne = annot.getParent(CtNamedElement.class);
        if( ne == null ) {
            SourcePosition sp = annot.getPosition();
            return sp.toString();
        }
        else {
            String s = SpoonHelper.toEclipseClickableString(ne);
            return s;
        }
    }
    
    /**
     * Return the list of fields holding references to components. Only retain
     * the references which are bound. Typically these are the fields which are
     * annotated with, for example in the case of Fraclet,
     * <code>@Requires</code>. But this can also be fields which are used by,
     * for example in the case of SCA, <code>@Reference</code> annotated setter
     * methods.
     * 
     * @param umclass  the component class
     * @return         the fields holding references to components
     */
    protected List<UMField<?>>
        getBoundReferenceFields( UMClass<?> umclass ) {
        
        List<UMField<?>> umfields = umclass.getUMFields();
        List<UMField<?>> ret = new ArrayList<UMField<?>>();
        
        for (UMField<?> umfield : umfields) {
            if( isRequiredAnnotatedUMField(umfield) &&
                isRequiredAnnotatedUMFieldCardinalitySingleton(umfield) &&
                isRequiredInterfaceBound(umfield,umclass) ) {
                
                ret.add(umfield);
            }
        }
        
        return ret;
    }
    
    /**
     * Return <code>true</code> if the specified required field in the specified
     * component class is bound to a target component. Return <code>false</code>
     * otherwise.
     * 
     * @author Frdric Loiret <Frederic.Loiret@inria.fr>
     */
    private boolean isRequiredInterfaceBound(
        UMField<?> field, UMClass<?> umclass ) {
        
        String refName = getRequiredAnnotatedUMFieldName(field);
        if( refName.length() == 0 ) {
            refName = field.getSimpleName();
        }
        return isRequiredInterfaceBound(refName,umclass);
    }    

    /**
     * Return <code>true</code> if the specified reference name in the specified
     * component class is bound to a target component. Return <code>false</code>
     * otherwise.
     * 
     * @author Frdric Loiret <Frederic.Loiret@inria.fr>
     */
    protected boolean isRequiredInterfaceBound(
        String refName, UMClass<?> umclass ) {
        
        ComponentDesc cdesc = umclass.getComponentDesc();
        BindingDesc bdesc = cdesc.getBinding(refName);
        
        if (bdesc == null) {
            return false;
        }
        else {
            UMClass<?> umc = getTargetComponentUMClass(bdesc);
            boolean b = umc != null;
            return b;
        }
    }    
}
