/***
 * Juliac
 * Copyright (C) 2008-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt.ultramerge;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.objectweb.fractal.juliac.desc.ComponentDesc;

import spoon.reflect.declaration.CtClass;
import spoon.reflect.declaration.CtField;
import spoon.reflect.declaration.CtMethod;
import spoon.reflect.declaration.CtParameter;
import spoon.reflect.declaration.CtSimpleType;
import spoon.reflect.reference.CtFieldReference;
import spoon.reflect.reference.CtTypeReference;

/**
 * A descriptor for a component content class.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 1.3
 */
public class UMClass<T> {
    
    /** The component descriptor. */
    private ComponentDesc cdesc;

    /** The component content class. */
    private CtClass<T> ctclass;
    
    /** The fields declared and inherited by the component content class. */
    private List<UMField<?>> umfields = new ArrayList<UMField<?>>();

    /** The methods declared and inherited by the component content class. */
    private List<UMMethod<?>> ummethods = new ArrayList<UMMethod<?>>();
    
    /**
     * The inheritance hierarchy of the component content class. The first
     * element in the list is the direct super class, while the last one is the
     * root of the hierarchy.
     */
    private List<UMClass<?>> umsupers = new ArrayList<UMClass<?>>();
    
    /**
     * @param cdesc    the component descriptor
     * @param ctclass  the component content class
     * @param scopes
     *      the classes where field accesses and method invocations to the
     *      members of this component content class must be sought
     * @throws IllegalComponentCodeException
     * 		if the representation can not be built
     */
    public UMClass(
		ComponentDesc cdesc, CtClass<T> ctclass, List<CtClass<?>> scopes )
    throws IllegalComponentCodeException {
        
    	this.cdesc = cdesc;
        this.ctclass = ctclass;
        
        /*
         * Retrieve declared and inherited fields.
         */
        CtTypeReference<?> tref = ctclass.getReference();
        Collection<CtFieldReference<?>> fieldrefs = tref.getAllFields();
        for (CtFieldReference<?> fieldref : fieldrefs) {
            CtField<?> field = fieldref.getDeclaration();
            if( field == null ) {
                String qname = fieldref.getQualifiedName();
                String msg = "Can not access field declaration for: "+qname;
                throw new IllegalComponentCodeException(msg);
            }
            UMField<?> umfield = new UMField(field,scopes);
            umfields.add(umfield);            
        }
        
        /*
         * Retrieve declared and inherited methods.
         */
        Set<CtMethod<?>> methods = ctclass.getAllMethods();
        for (CtMethod<?> method : methods) {
            UMMethod<?> ummethod = new UMMethod(method,scopes);
            ummethods.add(ummethod);
        }
        
        /*
         * Retrieve the inheritance hierarchy.
         */
        CtClass<?> src = ctclass;
        CtTypeReference<?> supercltref = null;
        List<CtClass<?>> srcs = new ArrayList<CtClass<?>>();

        do {
            srcs.add(src);
            supercltref = src.getSuperclass();
            
            // supercltref is null if there is no super class
            if( supercltref != null ) {
                CtSimpleType<?> stype = supercltref.getDeclaration();
                if( stype == null ) {
                    String msg =
                    	"Can not access super class declaration "+
                        supercltref.getQualifiedName()+
                        " for: "+src.getQualifiedName();
                    throw new IllegalComponentCodeException(msg);
                }
                src = (CtClass<?>) stype;
            }
        }
        while( supercltref != null );

        // Create the UMClasses corresponding to the inheritance hierarchy
        // Skip ctclass (first element in the list)
        for (int i = 1; i < srcs.size(); i++) {
        	CtClass<?> ctc = srcs.get(i);
            UMClass<?> umsuper = new UMClass(cdesc,ctc,srcs);
            umsupers.add(umsuper);
        }
    }
    
    public ComponentDesc getComponentDesc() {
        return cdesc;
    }    

    public CtClass<?> getCtClass() {
        return ctclass;
    }
    
    public void add( CtField<?> field ) {
    	/*
    	 * Assume that, by default, the field will only be referenced by its
    	 * parent class.
    	 */
    	List<CtClass<?>> scopes = new ArrayList<CtClass<?>>();
    	scopes.add(ctclass);
    	UMField<?> umfield = new UMField(field,scopes);
        umfields.add(umfield);            
    }
    
    /**
     * Return the field whose name is specified.
     * Return <code>null</code> if there is no such field.
     */
    public UMField<?> getUMField( String name ) {
        for (UMField<?> umfield : umfields) {
            if( name.equals(umfield.getSimpleName()) ) {
                return umfield;
            }
        }
        return null;
    }
    
    public List<UMField<?>> getUMFields() {
        return umfields;
    }
    
    /**
     * Return the method whose name and parameter types are specified.
     * Return <code>null</code> if there is no such method.
     */
    public UMMethod<?> getUMMethod( String name, List<CtTypeReference<?>> paramTypes ) {
        for (UMMethod<?> ummethod : ummethods) {
            if( name.equals(ummethod.getSimpleName()) ) {
                CtMethod<?> ctmethod = ummethod.getCtMethod();
                List<CtParameter<?>> params = ctmethod.getParameters();
                if( paramTypes.size() == params.size() ) {
                    for (int i = 0; i < paramTypes.size(); i++) {
                        CtTypeReference<?> paramType = paramTypes.get(i);
                        CtTypeReference<?> param = params.get(i).getType();
                        if( ! paramType.getQualifiedName().equals(param.getQualifiedName()) ) {
                            return null;
                        }
                    }
                    return ummethod;
                }
            }
        }
        return null;
    }
    
    public List<UMMethod<?>> getUMMethods() {
        return ummethods;
    }
    
    public List<UMClass<?>> getUMSupers() {
    	return umsupers;
    }
    
    @Override
    public String toString() {
        return ctclass.getQualifiedName();
    }
}
