/***
 * Juliac
 * Copyright (C) 2008-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Frederic Loiret
 */

package org.objectweb.fractal.juliac.opt.ultramerge;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.api.SpoonSupportItf;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.desc.BindingDesc;
import org.objectweb.fractal.juliac.desc.ComponentDesc;

import spoon.reflect.Factory;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.factory.ClassFactory;

/**
 * Repository for {@link UMClass}es.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Frederic Loiret <Frederic.Loiret@inria.fr>
 * @since 1.3
 */
public class UMClasses {
    
	private Juliac jc;
	
    public UMClasses( Juliac jc ) {
    	this.jc = jc;
        clear();
    }
    
    // ----------------------------------------------------------------------
    // Public methods
    // ----------------------------------------------------------------------
    
    /**
     * Recursively create the list of {@link UMClass}es associated to
     * the specified {@link ComponentDesc} and to the {@link ComponentDesc}s
     * bound to the specified {@link ComponentDesc}. Return the list of created
     * {@link UMClass}es.
     */
    public List<UMClass<?>> createAll( ComponentDesc cdesc )
    throws IllegalComponentCodeException {
        List<UMClass<?>> list = new ArrayList<UMClass<?>>();
        createAll(cdesc,list);
        return list;
    }
    
    /**
     * Return the {@link UMClass} associated to the specified
     * fully-qualified name. Return <code>null</code> if the {@link
     * UMClass} hasn't been created.
     */
    public UMClass<?> get( String qname ) {
        return umclasses.get(qname);
    }    
    
    /**
     * Return the {@link UMClass} associated to the specified {@link
     * ComponentDesc}. Return <code>null</code> if the {@link UMClass}
     * hasn't been created.
     */
    public UMClass<?> get( ComponentDesc cdesc ) {
        return cdescs.get(cdesc);
    }
    
    /**
     * Return the {@link ComponentDesc} associated to the specified {@link
     * UMClass}. Return <code>null</code> if the {@link ComponentDesc}
     * is not registered.
     */
    public ComponentDesc get( UMClass<?> umclass ) {
        return compumclasses.get(umclass);
    }
    
    /**
     * Return all created {@link UMClass}es.
     */
    public Collection<UMClass<?>> getAllUMClasses() {
        return umclasses.values();
    }
    
    /**
     * Return the names of all component classes contained in the repository
     * including the names of their super classes.
     */
    public Set<String> getAllUMClassNames() {
    	
    	/*
    	 * Copy the values since the collection returned by values() throws
    	 * UnsupportedOperationException when invoking: umcs.addAll(supers).
    	 */
    	Collection<UMClass<?>> values = umclasses.values();
    	Set<UMClass<?>> umcs = new HashSet<UMClass<?>>();
    	umcs.addAll(values);
    	
    	Set<UMClass<?>> supers = new HashSet<UMClass<?>>();
    	for (UMClass<?> umc : umcs) {
			List<UMClass<?>> sup = umc.getUMSupers();
			supers.addAll(sup);
		}
    	umcs.addAll(supers);
    	
    	Set<String> names = new HashSet<String>();
    	for (UMClass<?> umc : umcs) {
    		CtClass<?> ctclass = umc.getCtClass();
    		String name = ctclass.getQualifiedName();
    		names.add(name);
		}
    	
    	return names;
    }
    
    
    // ----------------------------------------------------------------------
    // Implementation specific
    // ----------------------------------------------------------------------
    
    private void createAll( ComponentDesc cdesc, List<UMClass<?>> list )
    throws IllegalComponentCodeException {
        
        if( cdescs.containsKey(cdesc) ) {
            return;
        }
        
        UMClass<?> umclass = create(cdesc);
        list.add(umclass);

        /*
         * Iterate on bindings and recursively compute the umclasses referenced
         * by the bindings.
         */
        Collection<BindingDesc> bdescs = cdesc.getBindingDescs();
        for (BindingDesc bdesc : bdescs) {
            ComponentDesc srv = bdesc.getSrv();
            createAll(srv,list);
        }
    }
    
    /**
     * Create the {@link UMClass} associated to the specified {@link
     * ComponentDesc}. Return the existing {@link UMClass} if it
     * already exists. Return <code>null</code> if the specified {@link
     * ComponentDesc} is not associated to a class (this is the case for
     * composite components.)
     */
    private UMClass<?> create( ComponentDesc cdesc )
    throws IllegalComponentCodeException {
        
        if( cdescs.containsKey(cdesc) ) {
            return cdescs.get(cdesc);
        }
        
        String contentClassName = cdesc.getContentClassName();
        
        if( contentClassName == null ) {
            /*
             * Composite component.
             * Add it in the map to avoid infinite recursion in case of cycle
             * between composite components. Bug reported by Frederic.
             */
            cdescs.put(cdesc,null);
            return null;
        }

    	JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportItf spoon = jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        ClassFactory cf = factory.Class();
        CtClass<Object> ctclass = cf.get(contentClassName);
        if( ctclass == null ) {
            String msg = "CtClass not found exception: "+contentClassName;
            throw new IllegalComponentCodeException(msg);
        }
        
        List<CtClass<?>> scopes = new ArrayList<CtClass<?>>();
        scopes.add(ctclass);
        UMClass<Object> umclass =
            new UMClass<Object>(cdesc,ctclass,scopes);
        register(contentClassName,umclass,cdesc);        
        return umclass;
    }    

    
    // ----------------------------------------------------------------------
    // Repository for {@link UMClass}es.
    //
    // These data structures are redundant at some point but allow accessing
    // {@link UMClass}es and {@link ComponentDesc}s by different ways.
    // ----------------------------------------------------------------------
    
    public void clear() {
        umclasses = new HashMap<String,UMClass<?>>();
        cdescs = new HashMap<ComponentDesc,UMClass<?>>();
        compumclasses = new HashMap<UMClass<?>,ComponentDesc>();
    }
    
    private void register(
        String qname, UMClass<?> umclass, ComponentDesc cdesc ) {
        
        umclasses.put(qname,umclass);
        cdescs.put(cdesc,umclass);
        compumclasses.put(umclass,cdesc);
    }
    
    /**
     * The {@link UMClass}es.
     */
    private Map<String,UMClass<?>> umclasses;
    
    /**
     * The {@link UMClass}es indexed by their associated {@link
     * ComponentDesc}.
     */
    private Map<ComponentDesc,UMClass<?>> cdescs;
    
    /**
     * The {@link ComponentDesc}s indexed by their associated {@link
     * UMClass}.
     */
    private Map<UMClass<?>,ComponentDesc> compumclasses;
}
